/* clavier.c - GTK+ "Clavier" Widget -- based on clavier-0.1.3
 * Copyright (C) 1998 Simon Kgedal
 * Copyright (C) 1999-2019 Michael Krause
 * Copyright (C) 2020-2021 Yury Aliaev
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License in the file COPYING for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * created 1998-04-18 Simon Kgedal
 */

#include <math.h>
#include <stdio.h>

#include "clavier.h"
#include "colors.h"
#include "marshal.h"

/* Signals */
enum {
    CLAVIERKEY_PRESS,
    CLAVIERKEY_RELEASE,
    CLAVIERKEY_ENTER,
    CLAVIERKEY_LEAVE,
    LAST_SIGNAL
};

static guint clavier_signals[LAST_SIGNAL] = { 0 };

typedef void (*ClavierSignal1)(GtkObject* object,
    gint arg1,
    gpointer data);

/* Clavier Methods */
static void clavier_class_init(ClavierClass* class);
static void clavier_init(Clavier* clavier);

/* GtkObject Methods */
static void clavier_destroy(GtkObject* object);

/* GtkWidget Methods */
static void clavier_realize(GtkWidget* widget);
static gint clavier_expose(GtkWidget* widget, GdkEventExpose* event);
static void clavier_size_allocate(GtkWidget* widget, GtkAllocation* allocation);
static gint clavier_button_press(GtkWidget* widget, GdkEventButton* event);
static gint clavier_button_release(GtkWidget* widget, GdkEventButton* event);
static gint clavier_motion_notify(GtkWidget* widget, GdkEventMotion* event);
static gint clavier_leave_notify(GtkWidget* widget, GdkEventCrossing* event);

G_DEFINE_TYPE(Clavier, clavier, custom_drawing_get_type())

static void
clavier_class_init(ClavierClass* class)
{
    GObjectClass* object_class;
    GtkObjectClass* gtkobject_class;
    GtkWidgetClass* widget_class;

    object_class = (GObjectClass*)class;
    gtkobject_class = (GtkObjectClass*)class;
    widget_class = (GtkWidgetClass*)class;

    clavier_signals[CLAVIERKEY_PRESS] = g_signal_new("clavierkey_press", G_TYPE_FROM_CLASS(object_class),
        (GSignalFlags)G_SIGNAL_RUN_FIRST,
        G_STRUCT_OFFSET(ClavierClass, clavierkey_press),
        NULL, NULL,
        __marshal_VOID__INT_INT, G_TYPE_NONE, 2,
        G_TYPE_INT, G_TYPE_INT);

    clavier_signals[CLAVIERKEY_RELEASE] = g_signal_new("clavierkey_release", G_TYPE_FROM_CLASS(object_class),
        (GSignalFlags)G_SIGNAL_RUN_FIRST,
        G_STRUCT_OFFSET(ClavierClass, clavierkey_release),
        NULL, NULL,
        __marshal_VOID__INT_INT, G_TYPE_NONE, 2,
        G_TYPE_INT, G_TYPE_INT);

    clavier_signals[CLAVIERKEY_ENTER] = g_signal_new("clavierkey_enter", G_TYPE_FROM_CLASS(object_class),
        (GSignalFlags)G_SIGNAL_RUN_FIRST,
        G_STRUCT_OFFSET(ClavierClass, clavierkey_enter),
        NULL, NULL,
        g_cclosure_marshal_VOID__INT, G_TYPE_NONE, 1,
        G_TYPE_INT);

    clavier_signals[CLAVIERKEY_LEAVE] = g_signal_new("clavierkey_leave", G_TYPE_FROM_CLASS(object_class),
        (GSignalFlags)G_SIGNAL_RUN_FIRST,
        G_STRUCT_OFFSET(ClavierClass, clavierkey_leave),
        NULL, NULL,
        g_cclosure_marshal_VOID__INT, G_TYPE_NONE, 1,
        G_TYPE_INT);

    gtkobject_class->destroy = clavier_destroy;

    widget_class->realize = clavier_realize;
    widget_class->expose_event = clavier_expose;
    widget_class->size_allocate = clavier_size_allocate;

    widget_class->button_press_event = clavier_button_press;
    widget_class->button_release_event = clavier_button_release;
    widget_class->motion_notify_event = clavier_motion_notify;
    widget_class->leave_notify_event = clavier_leave_notify;
}

static void
clavier_init(Clavier* clavier)
{
    clavier->keylabels = NULL;

    clavier->key_start = 36;
    clavier->key_end = 96;
    clavier->black_key_height = 0.6;
    clavier->black_key_width = 0.64;
    clavier->prev_width = 0;

    clavier->is_pressed = FALSE;
    clavier->key_pressed = 0;

    clavier->key_info = NULL;
    clavier->key_info_size = 0;

    di_widget_configure(GTK_WIDGET(clavier));
}

static void
clavier_destroy(GtkObject* object)
{
    Clavier* clavier;

    g_return_if_fail(object != NULL);
    g_return_if_fail(IS_CLAVIER(object));

    clavier = CLAVIER(object);

    /* eventually free memory allocated for key info
   */

    if (clavier->key_info) {
        g_free(clavier->key_info);
    }

    if (GTK_OBJECT_CLASS(clavier_parent_class)->destroy)
        (*GTK_OBJECT_CLASS(clavier_parent_class)->destroy)(object);
}

/* checks if the given key is a black one
 */

static gboolean
is_black_key(gint key)
{
    switch (key % 12) {
    case 1:
    case 3:
    case 6:
    case 8:
    case 10:
        return (TRUE);

    default:
        return (FALSE);
    }
}

/* calculates how many white and black keys there are in a certain
 * interval on the keyboard 
 */

static void
calc_keys(gint start, gint end, gint* whites, gint* blacks)
{
    gint i;

    *blacks = 0;
    *whites = 0;

    if (end >= start) {
        for (i = start; i <= end; i++) {
            if (is_black_key(i))
                (*blacks)++;
        }

        *whites = end - start + 1 - *blacks;
    }
}

/* this draws a key like on a real keyboard, y'know, all white keys have
 * the same width and the black keys are a bit narrower.
 */

static void
calc_key_normal(Clavier* clavier, gint key, ClavierKeyInfo* key_info)
{
    GtkWidget* widget;
    gint blacks, whites, width;

    widget = GTK_WIDGET(clavier);

    width = widget->allocation.width;
    calc_keys(clavier->key_start, key - 1, &whites, &blacks);

    switch (key % 12) {
    case 1:
    case 3:
    case 6:
    case 8:
    case 10:

        /* black key */

        key_info->is_black = TRUE;
        key_info->upper_right_x = ((gfloat)whites + (clavier->black_key_width * 0.5)) * clavier->keywidth;

        break;

    case 0:
    case 2:
    case 5:
    case 7:
    case 9:

        /* the (whites):th white key from left or bottom, with short
       * right side
       */

        key_info->is_black = FALSE;
        key_info->upper_right_x = (whites + 1) * clavier->keywidth - (clavier->black_key_width * 0.5 * clavier->keywidth);
        key_info->lower_right_x = (whites + 1) * clavier->keywidth;

        break;

    case 4:
    case 11:

        /* the (whites):th white key from left or bottom, with long
       * right side
       */

        key_info->is_black = FALSE;
        key_info->upper_right_x = key_info->lower_right_x = (whites + 1) * clavier->keywidth;

        break;
    }

    /* we need to do some fixing for the rightmost key */

    if (key == clavier->key_end) {
        key_info->upper_right_x = key_info->lower_right_x = width - 1;
    }
}

static void
calc_key_info(Clavier* clavier)
{
    gint keys = clavier->key_end - clavier->key_start + 1;
    gint i, prev = 0, pprev = 0;

    if (!clavier->key_info) {
        clavier->key_info = g_malloc(sizeof(ClavierKeyInfo) * keys);
    } else if (clavier->key_info_size != keys) {
        clavier->key_info = g_realloc(clavier->key_info, sizeof(ClavierKeyInfo) * keys);
    }

    for (i = 0; i < keys; i++) {
        ClavierKeyInfo* ki = &clavier->key_info[i];

        calc_key_normal(clavier,
            clavier->key_start + i, ki);

        if (!ki->is_black) {
            ki->width = ki->lower_right_x - prev;
            prev = ki->lower_right_x;
        } else
            ki->width = ki->upper_right_x - pprev;
        pprev = ki->upper_right_x;

        /* fix so draw_key draws lower rectangle correctly */

        if (i > 0 && (ki->is_black)) {
            clavier->key_info[i].lower_right_x = clavier->key_info[i - 1].lower_right_x;
        }
    }
}

static void
clavier_draw_label(Clavier* clavier,
    int keynum, gboolean pressed)
{
    if (clavier->keylabels) {
        ClavierKeyInfo* this = &(clavier->key_info[keynum]);
        gchar buf[4];
        gint8 label = clavier->keylabels[keynum];
        gboolean is_black = clavier->key_info[keynum].is_black;
        gint len;
        const gint black_pos =
            GTK_WIDGET(clavier)->allocation.height * clavier->black_key_height -
            clavier->fonth;
        const gint white_pos =
            MIN(GTK_WIDGET(clavier)->allocation.height * clavier->black_key_height + clavier->fonth,
            GTK_WIDGET(clavier)->allocation.height - clavier->fonth);

        if (label > 0xf) {
            len = sprintf(buf, "%x", label >> 4);
            di_draw_text(clavier->layout,
                colors_get_color(is_black ? (pressed ? COLOR_TEXT_BLACK_PRESSED : COLOR_TEXT_BLACK)
                                          : (pressed ? COLOR_TEXT_WHITE_PRESSED : COLOR_TEXT_WHITE)),
                buf, len, (is_black ? this->upper_right_x : this->lower_right_x)
                    - ((this->width + clavier->fontw) >> 1),
                (is_black ? black_pos : white_pos) - clavier->fonth);
        }
        len = sprintf(buf, "%x", label & 0xf);
        di_draw_text(clavier->layout,
            colors_get_color(is_black ? (pressed ? COLOR_TEXT_BLACK_PRESSED : COLOR_TEXT_BLACK)
                                      : (pressed ? COLOR_TEXT_WHITE_PRESSED : COLOR_TEXT_WHITE)),
            buf, len, (is_black ? this->upper_right_x : this->lower_right_x)
                - ((this->width + clavier->fontw) >> 1),
            is_black ? black_pos : white_pos);
    }
}

static void clavier_size_allocate(GtkWidget* widget,
    GtkAllocation* allocation)
{
    g_return_if_fail(widget != NULL);
    g_return_if_fail(IS_CLAVIER(widget));
    g_return_if_fail(allocation != NULL);

    widget->allocation.x = allocation->x;
    widget->allocation.y = allocation->y;
    widget->allocation.width = allocation->width;
    widget->allocation.height = allocation->height;

    if (gtk_widget_get_realized(widget)) {
        gdk_window_move_resize(widget->window, allocation->x, allocation->y,
            allocation->width, allocation->height);
    }

  (*GTK_WIDGET_CLASS(clavier_parent_class)->size_allocate)(widget, allocation);
}

/* main drawing function
 */

static gint
clavier_expose(GtkWidget* widget, GdkEventExpose* event)
{
    Clavier* clavier;
    int i;
    gint keys, whitekeys, blackkeys;
    gint width, height;
    ClavierKeyInfo first = { 0, 0 }, *prev;
    DIDrawable drw = custom_drawing_get_drawable(CUSTOM_DRAWING(widget));

    g_return_val_if_fail(widget != NULL, FALSE);
    g_return_val_if_fail(IS_CLAVIER(widget), FALSE);
    g_return_val_if_fail(event != NULL, FALSE);

    clavier = CLAVIER(widget);
    di_layout_set_drawable(clavier->layout,
        custom_drawing_get_gdk_drawable(CUSTOM_DRAWING(widget)));

    width = widget->allocation.width;
    height = widget->allocation.height;

    keys = clavier->key_end - clavier->key_start + 1;

    di_draw_rectangle(drw, clavier->white_gc, TRUE, 0, 0, width, height);

    /* To speed up a little bit */

    if (clavier->prev_width != width) {
        clavier->prev_width = width;
        calc_keys(clavier->key_start, clavier->key_end, &whitekeys, &blackkeys);
        clavier->keywidth = (gfloat)width / (gfloat)whitekeys;
        calc_key_info(clavier);
    }

    /*  first.upper_right_x = first.lower_right_x = 0;
    */
    prev = &first;

    for (i = 0; i < keys; i++) {
        gint x1, y1, x2, y2;

        if (clavier->key_info[i].is_black) {
            x1 = prev->upper_right_x + 1;
            y1 = 1;

            x2 = clavier->key_info[i].upper_right_x - prev->upper_right_x - 1;
            y2 = height * clavier->black_key_height;

            di_draw_rectangle(drw,
                clavier->black_gc,
                TRUE,
                x1, y1, x2, y2);
        } else {
            if (i != (keys - 1)) {
                x1 = clavier->key_info[i].lower_right_x;
                y1 = 1;
                x2 = x1;
                y2 = height;

                di_draw_line(drw,
                    clavier->black_gc,
                    x1, y1, x2, y2);
            }
        }

        prev = &clavier->key_info[i];

        clavier_draw_label(clavier, i, FALSE);
    }

    return (*GTK_WIDGET_CLASS(clavier_parent_class)->expose_event)(widget, event);
}

/* Draw a key pressed or unpressed
 */

static void
draw_key(Clavier* clavier, gint key, gboolean pressed)
{
    ClavierKeyInfo first = { 0, 0 }, *prev, *this;
    gint keynum;
    gint x1, y1, x2, y2, height;
    GtkWidget* widget = GTK_WIDGET(clavier);
    DIGC gc;
    DIDrawable drw = custom_drawing_get_drawable(CUSTOM_DRAWING(widget));
    GdkEventExpose event;

    di_layout_set_drawable(clavier->layout,
        custom_drawing_get_gdk_drawable(CUSTOM_DRAWING(widget)));
    height = widget->allocation.height;

    keynum = key - clavier->key_start;

    this = &(clavier->key_info[keynum]);
    prev = (keynum > 0) ? &(clavier->key_info[keynum - 1]) : &first;

    gc = this->is_black ? (pressed ? clavier->black_p_gc : clavier->black_gc)
        : (pressed ? clavier->white_p_gc : clavier->white_gc);

    /* draw upper part */

    x1 = prev->upper_right_x + 1;
    y1 = 1;
    x2 = this->upper_right_x - prev->upper_right_x - 1;
    y2 = height * clavier->black_key_height;

    di_draw_rectangle(drw, gc, TRUE, x1, y1, x2, y2);

    if (!clavier->key_info[keynum].is_black) {
        /* draw lower part */

        x1 = prev->lower_right_x + 1;
        y1 = height * clavier->black_key_height + 1;
        x2 = this->lower_right_x - prev->lower_right_x - 1;
        y2 = height * (1.0 - clavier->black_key_height) - 1;

        di_draw_rectangle(drw, gc, TRUE, x1, y1, x2, y2);
    }

    clavier_draw_label(clavier, keynum, pressed);

    /* Emulating exposure event to force redrawing from the image buffer */
    event.type = GDK_EXPOSE;
    event.window = widget->window;
    event.area.x = widget->allocation.x + x1;
    event.area.y = widget->allocation.y + 1;
    event.area.width = x2;
    event.area.height = widget->allocation.height - 1;
    (*GTK_WIDGET_CLASS(clavier_parent_class)->expose_event)(widget, &event);
}

/* See which key is drawn at x, y
 */

static gint
which_key(Clavier* clavier, gint x, gint y)
{
    gint i, keys, height;

    height = GTK_WIDGET(clavier)->allocation.height;

    keys = clavier->key_end - clavier->key_start + 1;

    if (y > (height * clavier->black_key_height)) {
        /* check lower part */

        for (i = 0; i < keys; i++) {
            if (x < clavier->key_info[i].lower_right_x) {
                return clavier->key_start + i;
            }
        }
    } else {
        /* check upper part */

        for (i = 0; i < keys; i++) {
            if (x < clavier->key_info[i].upper_right_x) {
                return (clavier->key_start + i);
            }
        }
    }

    /* so it must be the rightmost key */

    return (clavier->key_end);
}

static void
press_key(Clavier* clavier, gint key, gint button)
{
    clavier->is_pressed = TRUE;
    clavier->key_pressed = key;
    clavier->button = button;

    g_signal_emit(G_OBJECT(clavier), clavier_signals[CLAVIERKEY_PRESS], 0, key, button);

    /*  printf("press: %i\n", key); */
}

static void
release_key(Clavier* clavier, gint button)
{
    clavier->is_pressed = FALSE;

    g_signal_emit(G_OBJECT(clavier), clavier_signals[CLAVIERKEY_RELEASE], 0,
        clavier->key_pressed, button);

    /*  printf ("release: %i\n", clavier->key_pressed); */
}

/* events
 */

static gint
clavier_button_press(GtkWidget* widget, GdkEventButton* event)
{
    Clavier* clavier;
    gint key;

    g_return_val_if_fail(widget != NULL, FALSE);
    g_return_val_if_fail(IS_CLAVIER(widget), FALSE);
    g_return_val_if_fail(event != NULL, FALSE);

    clavier = CLAVIER(widget);

    key = which_key(clavier, (gint)event->x, (gint)event->y);

    press_key(clavier, key, event->button);

    /* add pointer grab - hmm, doesn't seem to work the way i'd like :( */

    gtk_grab_add(widget);

    return FALSE;
}

static gint
clavier_button_release(GtkWidget* widget, GdkEventButton* event)
{
    Clavier* clavier;

    g_return_val_if_fail(widget != NULL, FALSE);
    g_return_val_if_fail(IS_CLAVIER(widget), FALSE);
    g_return_val_if_fail(event != NULL, FALSE);

    clavier = CLAVIER(widget);

    release_key(clavier, event->button);

    gtk_grab_remove(widget);

    return FALSE;
}

static gint
clavier_motion_notify(GtkWidget* widget, GdkEventMotion* event)
{
    Clavier* clavier;
    gint key;

    g_return_val_if_fail(widget != NULL, FALSE);
    g_return_val_if_fail(IS_CLAVIER(widget), FALSE);
    g_return_val_if_fail(event != NULL, FALSE);

    clavier = CLAVIER(widget);

    key = which_key(clavier, (gint)event->x, (gint)event->y);

    if (key != clavier->key_entered) {
        if (clavier->key_entered != -1) {
            g_signal_emit(G_OBJECT(clavier),
                clavier_signals[CLAVIERKEY_LEAVE], 0, clavier->key_entered);
        }
        clavier->key_entered = key;
        g_signal_emit(G_OBJECT(clavier),
            clavier_signals[CLAVIERKEY_ENTER], 0, key);
    }

    if (clavier->is_pressed) {
        if (key != clavier->key_pressed) {
            release_key(clavier, clavier->button);
            press_key(clavier, key, clavier->button);
        }
    }

    return FALSE;
}

static gint
clavier_leave_notify(GtkWidget* widget,
    GdkEventCrossing* event)
{
    Clavier* clavier = CLAVIER(widget);

    if (clavier->key_entered != -1) {
        g_signal_emit(G_OBJECT(clavier),
            clavier_signals[CLAVIERKEY_LEAVE], 0, clavier->key_entered);
        clavier->key_entered = -1;
    }

    return FALSE;
}

GtkWidget*
clavier_new(const gchar* font)
{
    Clavier* clavier;

    clavier = g_object_new(clavier_get_type(), NULL);
    clavier->font = font;

    return GTK_WIDGET(clavier);
}

static inline void
clavier_update_layout(Clavier* clavier)
{
    di_layout_set_font(clavier->layout, clavier->font);
    /* let's just hope this is a non-proportional font */
    di_layout_get_pixel_size(clavier->layout, "0", -1, &clavier->fontw, &clavier->fonth);
}

static void
clavier_realize(GtkWidget* widget)
{
    GdkWindowAttr attributes;
    Clavier* clavier;
    gint attributes_mask;

    g_return_if_fail(widget != NULL);
    g_return_if_fail(IS_CLAVIER(widget));

    clavier = CLAVIER(widget);
    gtk_widget_set_realized(widget, TRUE);

    attributes.window_type = GDK_WINDOW_CHILD;
    attributes.x = widget->allocation.x;
    attributes.y = widget->allocation.y;
    attributes.width = widget->allocation.width;
    attributes.height = widget->allocation.height;
    attributes.wclass = GDK_INPUT_OUTPUT;
    attributes.visual = gtk_widget_get_visual(widget);
    attributes.colormap = gtk_widget_get_colormap(widget);
    attributes.event_mask = gtk_widget_get_events(widget) | GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK | GDK_POINTER_MOTION_MASK | GDK_LEAVE_NOTIFY_MASK;

    attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;

    widget->window = gdk_window_new(gtk_widget_get_parent_window(widget),
        &attributes, attributes_mask);
    gdk_window_set_user_data(widget->window, clavier);

    widget->style = gtk_style_attach(widget->style, widget->window);
    gtk_style_set_background(widget->style, widget->window, GTK_STATE_NORMAL);

    clavier->white_gc = colors_get_gc(COLOR_KEY_WHITE);
    colors_add_widget(COLOR_KEY_WHITE, widget);
    clavier->black_gc = colors_get_gc(COLOR_KEY_BLACK);
    clavier->white_p_gc = colors_get_gc(COLOR_KEY_WHITE_PRESSED);
    clavier->black_p_gc = colors_get_gc(COLOR_KEY_BLACK_PRESSED);
    clavier->white_tp_gc = colors_get_gc(COLOR_TEXT_WHITE_PRESSED);
    clavier->black_tp_gc = colors_get_gc(COLOR_TEXT_BLACK_PRESSED);

    clavier->layout = di_layout_new(widget);
    clavier_update_layout(clavier);
    (*GTK_WIDGET_CLASS(clavier_parent_class)->realize)(widget);
}

/* these routines should most often be mapped to the press and release
 * signals, but not all applications would want it that way
 */

void clavier_press(Clavier* clavier, gint key)
{
    /* TODO: should keep track of what keys are pressed so that the
   * whole widget can be correctly redrawn
   */

    if (key < clavier->key_start || key > clavier->key_end)
        return;

    draw_key(clavier, key, TRUE);
}

void clavier_release(Clavier* clavier, gint key)
{
    if (key < clavier->key_start || key > clavier->key_end)
        return;

    draw_key(clavier, key, FALSE);
}

/* attribute setting 
 */

void clavier_set_range(Clavier* clavier, gint start, gint end)
{
    g_return_if_fail(start >= 0 && start <= 127);
    g_return_if_fail(start >= 0 && start <= 127);
    g_return_if_fail(end >= start);

    clavier->key_start = start;
    clavier->key_end = end;
}

void clavier_set_key_labels(Clavier* clavier,
    gint8* labels)
{
    clavier->keylabels = labels;
    gtk_widget_queue_draw(GTK_WIDGET(clavier));
}

void clavier_set_font(Clavier* clavier,
    const gchar* font)
{
    clavier->font = font;
    clavier_update_layout(clavier);
    gtk_widget_queue_draw(GTK_WIDGET(clavier));
}
